// Original by David Eck, adapted by Alex Gerdes

import java.util.ArrayList;
import java.util.List;

public class DivideAndConquer {
  private static final int END = 10000000;  // Should be divisible by N_THREADS
  private static final int N_THREADS = 20;   // Number of threads
  private static int total = 0;             // total number of primes

  public static void main(String[] args) {
    int n = Runtime.getRuntime().availableProcessors();
    System.out.println("Your computer has " + n + " available processor(s).\n");

    countPrimesWithThreads(N_THREADS);
  }

  synchronized private static void addToTotal(int x) {  // Critical section, only one thread should do this at a time
    total = total + x;
    System.out.println(total + " primes found so far.");
  }

  private static void countPrimesWithThreads(int numberOfThreads) {
    Workers workers = divideWork(numberOfThreads);

    System.out.println( "\nCounting primes between 1 and " + END
                      + " using " + numberOfThreads + " threads...\n");

    long elapsedTime = measureTime(() -> {workers.start(); workers.joinAll();});

    System.out.println("\nThe number of primes is " + total + ".");
    System.out.println("\nTotal elapsed time:  " + (elapsedTime / 1000.0) + " seconds.\n");
  }

  private static Workers divideWork(int numberOfThreads) {
    int increment = END / numberOfThreads;
    Workers workers = new Workers();

    for (int i = 0; i < numberOfThreads; i++) {
      final int min = i * increment + 1;
      final int max = (i + 1) * increment;
      workers.add(() -> addToTotal(Prime.countPrimes(min, max)));
    }

    return workers;
  }

  private static long measureTime(Runnable runnable) {
    long startTime = System.currentTimeMillis();
    runnable.run();
    return System.currentTimeMillis() - startTime;
  }
}

class Workers extends Thread {
  List<Thread> pool = new ArrayList<>();  // Create a pool of workers

  public void add(Runnable run) {
    pool.add(new Thread(run));
  }

  @Override
  public void start() {
    for (Thread thread : pool)  // Start threads
      thread.start();
  }

  public void joinAll() {
    for (Thread thread : pool) {  // Wait for all threads to finish
      while (thread.isAlive()) {
        try {
          thread.join();
        } catch (InterruptedException e) {
        }
      }
    }
  }
}

class Prime {
  // Count number of primes between min and max
  static int countPrimes(int min, int max) {
    int count = 0;

    for (int i = min; i <= max; i++)
      if (isPrime(i))
        count++;

    System.out.println("There are " + count + " primes between " + min + " and " + max);

    return count;
  }

  static boolean isPrime(int x) {
    int top = (int) Math.sqrt(x);

    for (int i = 2; i <= top; i++)
      if (x % i == 0)
        return false;

    return true;
  }
}